use super::command::Command;
use super::mpv_framed::MpvFramed;
use bytes::BytesMut;
use std::borrow::BorrowMut;
use std::collections::hash_map::Entry;
use std::collections::HashMap;
use std::io::Read;
use std::os::unix::prelude::OsStrExt;
use std::sync::{Arc, Mutex};
use std::{collections::BTreeMap, path::Path};
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use tokio::net::unix::OwnedWriteHalf;
use tokio::net::UnixStream;
use tokio::sync::oneshot;

type CommandDb = Arc<Mutex<HashMap<u64, oneshot::Sender<serde_json::Value>>>>;

pub struct MpvCtl {
    socket: OwnedWriteHalf,
    next_request_id: u64,
    in_flight_requests: CommandDb,
}

impl MpvCtl {
    pub async fn connect<P: AsRef<Path>>(
        socket_path: P,
    ) -> Result<Self, Box<dyn std::error::Error>> {
        let socket = UnixStream::connect(socket_path).await?;
        let (socket_read, socket_write) = socket.into_split();
        let mut framed_read = MpvFramed::new(socket_read);
        let db: CommandDb = Arc::new(Mutex::new(HashMap::new()));
        let read_db_handle = db.clone();

        tokio::spawn(async move {
            loop {
                match Self::read_loop(&mut framed_read, &read_db_handle).await {
                    Ok(_) => (),
                    Err(e) => {
                        eprintln!("Error in read loop: {}", e);
                    }
                };
            }
        });

        Ok(MpvCtl {
            socket: socket_write,
            next_request_id: 0,
            in_flight_requests: db,
        })
    }

    async fn read_loop(
        framed_read: &mut MpvFramed,
        db: &CommandDb,
    ) -> Result<(), Box<dyn std::error::Error>> {
        match framed_read.read_frame().await {
            Ok(Some(frame)) => {
                println!("Read {}", frame);
                // get the request id and push the result into the channel
                let reqid = {
                    let obj = match &frame {
                        serde_json::Value::Object(obj) => obj,
                        _ => {
                            return Err("Got back a json value that wasn't an object.".into());
                        }
                    };
                    match obj.get("request_id") {
                        Some(serde_json::Value::Number(reqid)) if reqid.is_u64() => {
                            reqid.as_u64().unwrap()
                        }
                        _ => {
                            return Err("Unrecognized request_id".into());
                        }
                    }
                };

                {
                    let mut db_handle = db.lock().unwrap();
                    if let Entry::Occupied(o) = (*db_handle).entry(reqid) {
                        o.remove()
                            .send(frame)
                            .map_err(|e| "Failed to send frame".to_string())?;
                    } else {
                        return Err("No entry found for request id".into());
                    }
                }
            }
            Ok(None) => {
                return Ok(());
            }
            Err(e) => {
                return Err("Ran into a problem".into());
            }
        }
        Ok(())
    }

    fn grab_request_id(&mut self) -> u64 {
        let ret = self.next_request_id;
        self.next_request_id += 1;
        ret
    }

    async fn send_command(
        &mut self,
        cmd: Command,
    ) -> Result<oneshot::Receiver<serde_json::Value>, Box<dyn std::error::Error>> {
        let serialized = serde_json::to_string(&cmd)? + "\n";
        print!("Sending {}", serialized);
        let (response_tx, response_rx) = oneshot::channel::<serde_json::Value>();
        match cmd.get_request_id() {
            Some(reqid) => {
                self.in_flight_requests
                    .lock()
                    .unwrap()
                    .insert(*reqid, response_tx);
            }
            None => (),
        };

        self.socket.write_all(serialized.as_bytes()).await?;

        Ok(response_rx)
    }

    pub async fn get_client_name(&mut self) -> Result<String, Box<dyn std::error::Error>> {
        let request_id = self.grab_request_id();
        let cmd = Command::new(vec!["client_name".to_string()], Some(request_id));
        let cmd_result = self.send_command(cmd).await?;
        let result = cmd_result.await?;
        if let serde_json::Value::Object(obj) = result {
            let client_name = match obj.get("data") {
                Some(serde_json::Value::String(client_name)) => client_name,
                _ => {
                    return Err("Failed to get client name".into());
                }
            };
            return Ok(client_name.to_owned());
        } else {
            return Err("Did not get back an object".into());
        }
    }

    pub async fn play_video<P: AsRef<Path>>(
        &mut self,
        path: P,
    ) -> Result<(), Box<dyn std::error::Error>> {
        let request_id = self.grab_request_id();
        let file_path = path.as_ref().to_str().unwrap().to_string();
        let cmd = Command::new(vec!["loadfile".to_string(), file_path], Some(request_id));
        let cmd_result = self.send_command(cmd).await?;
        let result = cmd_result.await?;
        if let serde_json::Value::Object(obj) = result {
            match obj.get("error") {
                Some(status) if status == "success" => (),
                _ => {
                    return Err("Non-success status".into());
                }
            };
            return Ok(());
        } else {
            return Err("Did not get back an object".into());
        }
    }
}
